﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Data.OleDb;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using C1.Win.C1SpellChecker;

namespace C1DBSpellChecker
{
    public partial class Form1 : Form
    {
        //------------------------------------------------------------------------------
        #region ** fields

        // current connection string and corresponding schema
        string _connString;
        OleDbSchema _schema;

        // updating tree checkboxes
        bool _updatingTree;

        // max number of error rows to show in each session
        const int MAX_ERROR_ROWS = 500;

        #endregion

        //------------------------------------------------------------------------------
        #region ** ctor

        public Form1()
        {
            InitializeComponent();

            // make combo owner-drawn
            var cmb = _cmbConnString.ComboBox;
            cmb.DrawMode = DrawMode.OwnerDrawFixed;
            cmb.DrawItem += cmb_DrawItem;
            _toolStrip_SizeChanged(_toolStrip, EventArgs.Empty);
        }

        #endregion

        //------------------------------------------------------------------------------
        #region ** object model

        /// <summary>
        /// Gets or sets the connection string.
        /// </summary>
        public string ConnectionString
        {
            get { return _connString; }
            set
            {
                if (value != ConnectionString)
                {
                    // this may take a while
                    Cursor.Current = Cursors.WaitCursor;

                    // look for item in the list
                    var items = _cmbConnString.Items;
                    var index = items.IndexOf(value);

                    // try getting schema for the new connection string
                    _schema = OleDbSchema.GetSchema(value);

                    // handle good connection strings
                    if (_schema != null)
                    {
                        // add good values to the list
                        if (index < 0)
                        {
                            items.Insert(0, value);
                        }
                        else if (index > 0)
                        {
                            items.RemoveAt(index);
                            items.Insert(0, value);
                        }

                        // trim list
                        while (items.Count > _cmbConnString.MaxDropDownItems)
                        {
                            items.RemoveAt(items.Count - 1);
                        }
                    }
                    else // handle bad connection strings
                    {
                        // remove from list
                        if (index >= 0)
                        {
                            items.RemoveAt(index);
                        }

                        // do not store bad values
                        value = string.Empty;
                    }

                    // save new value
                    _connString = value;

                    // show new value in combo box and table tree
                    _cmbConnString.Text = value;
                    UpdateTableTree();
                    Cursor.Current = Cursors.Default;
                }
            }
        }
        /// <summary>
        /// Gets a SQL statement that corresponds to the element that
        /// is currently selected (table, view, stored procedure, or
        /// explicit sql statement).
        /// </summary>
        public string SelectStatement
        {
            get
            {
                var nd = _treeTables.SelectedNode;
                return nd == null || nd.Tag == null || _schema == null
                    ? string.Empty
                    : _schema.GetSelectStatement(nd.Tag as DataTable);
            }
        }

        #endregion

        //------------------------------------------------------------------------------
        #region ** event handlers

        // form loaded: load app settings
        protected override void OnLoad(EventArgs e)
        {
            // load recently used connection strings
            var mru = Properties.Settings.Default.RecentConnections;
            if (mru != null)
            {
                foreach (string connString in mru)
                {
                    _cmbConnString.Items.Add(connString);
                }
            }

            // load spell-checker options
            var ignore = Properties.Settings.Default.Ignore;
            if (ignore > -1)
            {
                _spell.Options.Ignore = (IgnoreOptions)ignore;
            }
            var userDict = Properties.Settings.Default.UserDict;
            if (!string.IsNullOrEmpty(userDict))
            {
                _spell.UserDictionary.FileName = userDict;
            }
            var mainDict = Properties.Settings.Default.MainDict;
            if (!string.IsNullOrEmpty(mainDict))
            {
                _spell.MainDictionary.FileName = mainDict;
            }

            // default handling
            base.OnLoad(e);
        }

        // form closing: save app settings
        protected override void OnFormClosing(FormClosingEventArgs e)
        {
            // save recently used connection strings
            var mru = new System.Collections.Specialized.StringCollection();
            foreach (string item in _cmbConnString.Items)
            {
                mru.Add(item);
            }
            Properties.Settings.Default.RecentConnections = mru;

            // save spell-checker options
            Properties.Settings.Default.Ignore = (int)_spell.Options.Ignore;
            Properties.Settings.Default.UserDict = _spell.UserDictionary.FileName;
            Properties.Settings.Default.MainDict = _spell.MainDictionary.FileName;

            // save settings
            Properties.Settings.Default.Save();

            // default handling
            base.OnFormClosing(e);
        }

        // validating combo: apply new connection string
        void _cmbConnString_Validating(object sender, CancelEventArgs e)
        {
            ConnectionString = _cmbConnString.Text;
        }

        // new connection picked: apply new connection string
        void _cmbConnString_SelectedIndexChanged(object sender, EventArgs e)
        {
            ConnectionString = _cmbConnString.Text;
            _treeTables.Focus();
        }

        // user pressed enter: apply new connection string
        void _cmbConnString_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == 13)
            {
                e.Handled = true;
                ConnectionString = _cmbConnString.Text;
                _treeTables.Focus();
            }
        }

        // pick a new connection
        void _btnConnPicker_Click(object sender, EventArgs e)
        {
            // release mouse capture to avoid wait cursor
            _toolStrip.Capture = false;

            // get starting connection string
            // (if empty or no provider, start with SQL source as default)
            string connString = _cmbConnString.Text;
            if (string.IsNullOrEmpty(connString) || connString.IndexOf("provider=", StringComparison.OrdinalIgnoreCase) < 0)
            {
                connString = "Provider=SQLOLEDB.1;";
            }

            // let user change it
            connString = PromptConnectionString(connString);
            ConnectionString = connString;
            _treeTables.Focus();
        }

        // trim items in combo using ellipsis (they're very long)
        void cmb_DrawItem(object sender, DrawItemEventArgs e)
        {
            var fmt = new StringFormat();
            fmt.LineAlignment = StringAlignment.Center;
            fmt.Trimming = StringTrimming.EllipsisPath;

            var text = (string)_cmbConnString.Items[e.Index];
            text = TrimConnectionString(text);

            var brush = (e.State & DrawItemState.Selected) != 0
                ? SystemBrushes.HighlightText
                : SystemBrushes.WindowText;

            e.DrawBackground();
            e.Graphics.DrawString(text, _cmbConnString.Font, brush, e.Bounds, fmt);
            e.DrawFocusRectangle();
        }

        // stretch combo to fill tool strip
        void _toolStrip_SizeChanged(object sender, EventArgs e)
        {
            int width = _toolStrip.ClientRectangle.Width - 5;
            foreach (ToolStripItem item in _toolStrip.Items)
            {
                if (item != _cmbConnString)
                {
                    width -= item.Width - item.Margin.Left - item.Margin.Right;
                }
            }
            width = Math.Max(100, Math.Min(500, width));
            _cmbConnString.Width = width;
        }

        // handle check boxes
        void _treeTables_AfterCheck(object sender, TreeViewEventArgs e)
        {
            if (_updatingTree)
            {
                return;
            }

            // start updating...
            _updatingTree = true;

            // get node that was clicked
            var n = e.Node;

            // clicked on table node
            if (n.Tag is DataTable)
            {
                // apply check state to all child nodes (fields)
                SetCheck(n, n.Checked);

                // if this table is checked, uncheck all other tables
                if (n.Checked)
                {
                    foreach (TreeNode c in n.TreeView.Nodes)
                    {
                        if (c != n)
                        {
                            SetCheck(c, false);
                        }
                    }
                }
            }

            // clicked on column node
            if (n.Tag is DataColumn)
            {
                // update parent node state
                bool check = false;
                foreach (TreeNode c in n.Parent.Nodes)
                {
                    check |= c.Checked;
                }
                n.Parent.Checked = check;

                // if this node is checked, uncheck all other tables
                if (n.Checked)
                {
                    foreach (TreeNode c in n.TreeView.Nodes)
                    {
                        if (c != n.Parent)
                        {
                            SetCheck(c, false);
                        }
                    }
                }
            }

            // done updating...
            _updatingTree = false;
        }
        void SetCheck(TreeNode n, bool check)
        {
            n.Checked = check;
            foreach (TreeNode c in n.Nodes)
            {
                c.Checked = check;
            }
        }

        // done...
        void _btnCancel_Click(object sender, EventArgs e)
        {
            Close();
        }

        // spell-check selection
        void _btnSpell_Click(object sender, EventArgs e)
        {
            // get list of checked columns
            DataTable table = null;
            var columns = new List<string>();
            foreach (TreeNode n in _treeTables.Nodes)
            {
                foreach (TreeNode c in n.Nodes)
                {
                    if (c.Checked)
                    {
                        var col = c.Tag as DataColumn;
                        columns.Add(col.ColumnName);
                        table = col.Table;
                    }
                }
            }

            // no table selected? we're done
            if (table == null)
            {
                Warning(
                    string.IsNullOrEmpty(ConnectionString)
                        ? Properties.Resources.SelectConnectionString
                        : Properties.Resources.SelectColumns);
                return;
            }

            // build DataAdapter with select and update commands
            var sql = _schema.GetSelectStatement(table);
            var da = new OleDbDataAdapter(sql, ConnectionString);
            var bld = new OleDbCommandBuilder(da);
            da.UpdateCommand = bld.GetUpdateCommand();

            // get data table with spelling errors
            var dt = GetSpellingErrors(da, columns);
            if (dt.Rows.Count == 0)
            {
                MessageBox.Show(
                    this,
                    Properties.Resources.NoSpellingErrors,
                    Application.ProductName,
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Information);
                return;
            }

            // show spell-checking form
            using (var f = new SpellCheckerForm())
            {
                f.Font = Font;
                f.SpellChecker = _spell;
                f.DataTable = dt;
                f.DataAdapter = da;
                f.ShowDialog();
            }
        }

        #endregion

        //------------------------------------------------------------------------------
        #region ** implementation

        // prompt user for a connection string
        string PromptConnectionString(string connString)
        {
            try
            {
                // create objects we'll need
                var dlinks = new MSDASC.DataLinksClass();
                var conn = new ADODB.ConnectionClass();

                // sanity
                if (dlinks == null || conn == null)
                {
                    Warning(Properties.Resources.ErrDataLinks);
                    return connString;
                }

                // initialize object
                if (!string.IsNullOrEmpty(connString))
                {
                    conn.ConnectionString = connString;
                }

                // show connection picker dialog
                object obj = conn;
                dlinks.hWnd = (int)Handle;
                if (dlinks.PromptEdit(ref obj))
                {
                    connString = conn.ConnectionString;
                }
            }
            catch (Exception x)
            {
                Warning(Properties.Resources.ErrConnectionString, x.Message);
            }

            // done
            return connString;
        }

        // issue a warning
        void Warning(string format, params object[] args)
        {
            string msg = string.Format(format, args);
            MessageBox.Show(this, msg, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Warning);
        }

        // trim connection string for display
        string[] _keys = new string[] { "Provider", "Initial Catalog", "Data Source" };
        string TrimConnectionString(string text)
        {
            var sb = new StringBuilder();
            foreach (var item in text.Split(';'))
            {
                foreach (var key in _keys)
                {
                    if (item.IndexOf(key, StringComparison.InvariantCultureIgnoreCase) > -1)
                    {
                        if (sb.Length > 0)
                        {
                            sb.Append("...");
                        }
                        sb.Append(item.Split('=')[1].Trim());
                    }
                }
            }
            return sb.ToString();
        }

        // update table tree to reflect new connection string
        void UpdateTableTree()
        {
            // initialize table tree
            TreeNodeCollection nodes = _treeTables.Nodes;
            nodes.Clear();

            // populate using current schema
            if (_schema != null)
            {
                // populate the tree
                _treeTables.BeginUpdate();
                foreach (DataTable dt in _schema.Tables)
                {
                    if (_schema.GetTableType(dt) == TableType.Table)
                    {
                        // create new node, save table in tag property
                        var node = new TreeNode(dt.TableName);
                        node.Tag = dt;

                        // add string fields to node
                        foreach (DataColumn col in dt.Columns)
                        {
                            if (col.DataType == typeof(string))
                            {
                                var ndCol = node.Nodes.Add(col.ColumnName);
                                ndCol.Tag = col;
                            }
                        }

                        // add new node to the tree
                        if (node.Nodes.Count > 0)
                        {
                            nodes.Add(node);
                        }
                    }
                }

                // done
                _treeTables.Sort();
                _treeTables.EndUpdate();
            }
        }

        // get data table with the rows that contain spelling errors
        DataTable GetSpellingErrors(OleDbDataAdapter da, List<string> columns)
        {
            // this will take a while
            Cursor.Current = Cursors.WaitCursor;

            // create table and adapter
            var dt = new DataTable();

            // get table schema
            da.FillSchema(dt, SchemaType.Mapped);

            // make columns not being checked read-only
            foreach (DataColumn col in dt.Columns)
            {
                if (!columns.Contains(col.ColumnName))
                {
                    col.ReadOnly = true;
                }
            }

            // read rows with DataReader
            int errorCount = 0;
            int rowCount = 0;
            var cmd = da.SelectCommand;
            cmd.Connection.Open();
            using (var reader = cmd.ExecuteReader())
            {
                while (reader.Read())
                {
                    // read a row
                    rowCount++;
                    var dr = dt.NewRow();
                    foreach (DataColumn col in dt.Columns)
                    {
                        var index = col.Ordinal;
                        dr[index] = reader.GetValue(index);
                    }

                    // check for errors
                    bool hasErrors = false;
                    foreach (string col in columns)
                    {
                        var text = dr[col] as string;
                        if (!string.IsNullOrEmpty(text))
                        {
                            var errors = _spell.CheckText(text);
                            if (errors.Count > 0)
                            {
                                hasErrors = true;
                                errorCount += errors.Count;
                            }
                        }
                    }

                    // keep rows that have errors
                    if (hasErrors)
                    {
                        dt.Rows.Add(dr);
                        if (dt.Rows.Count >= MAX_ERROR_ROWS)
                        {
                            break;
                        }
                    }

                    // update status
                    if (rowCount % 500 == 0)
                    {
                        _lblStatus.Text = string.Format("{0:n0} rows, {1:n0} errors", rowCount, errorCount);
                        _lblStatus.Update();
                    }
                }
            }

            // table has no changes
            dt.AcceptChanges();

            // done
            Cursor.Current = Cursors.Default;
            _lblStatus.Text = string.Empty;
            return dt;
        }

        #endregion
    }
}
