﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Data.OleDb;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using C1.Win.C1FlexGrid;
using C1.Win.C1SpellChecker;

namespace C1DBSpellChecker
{
    public partial class SpellCheckerForm : Form
    {
        //---------------------------------------------------------------------------
        #region ** fields

        DataTable _dtErrors;
        OleDbDataAdapter _dataAdapter;
        CellStyle _errorStyle, _readOnlyStyle;
        C1SpellChecker _spell;

        #endregion

        //---------------------------------------------------------------------------
        #region ** ctor

        public SpellCheckerForm()
        {
            InitializeComponent();

            // configure flex grid
            _flex.Styles.Normal.WordWrap = true;
            _flex.AutoResize = false;

            // create style used to highlight errors
            _errorStyle = _flex.Styles.Add("error");
            _errorStyle.ForeColor = Color.DarkRed;

            // create style used show read-only columns
            _readOnlyStyle = _flex.Styles.Add("readOnly");
            _readOnlyStyle.BackColor = Color.WhiteSmoke;

            // keep track of changes
            _flex.CellChanged += _flex_CellChanged;
        }

        #endregion

        //---------------------------------------------------------------------------
        #region ** object model

        public C1SpellChecker SpellChecker
        {
            get { return _spell; }
            set { _spell = value; }
        }
        public DataTable DataTable
        {
            get { return _dtErrors; }
            set { _dtErrors = value; }
        }
        public OleDbDataAdapter DataAdapter
        {
            get { return _dataAdapter; }
            set { _dataAdapter = value; }
        }

        #endregion

        //---------------------------------------------------------------------------
        #region ** overrides

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            // update form caption
            Text = string.Format(Text, _dtErrors.TableName, _dtErrors.Rows.Count);

            // show fields with errors on grid
            _flex.DataSource = _dtErrors;

            // auto-size rows/cols
            _flex.Cols.MaxSize = 600;
            _flex.AutoSizeCols(4);
            _flex.AutoSizeRows();

            // apply styles to columns, move cursor to first editable cell
            _flex.Col = -1;
            for (int index = _flex.Cols.Fixed; index < _flex.Cols.Count; index++)
            {
                var col = _flex.Cols[index];
                if (col.AllowEditing)
                {
                    if (_flex.Col < 0)
                    {
                        _flex.Col = index;
                    }
                }
                else
                {
                    col.Style = _readOnlyStyle;
                }
            }
        }

        #endregion

        //---------------------------------------------------------------------------
        #region ** implementation

        // check that the user wants to leave without saving the changes
        bool OKToDiscardChanges()
        {
            if (HasChanges())
            {
                var dr = MessageBox.Show(this,
                    Properties.Resources.DiscardChanges,
                    Application.ProductName, 
                    MessageBoxButtons.YesNo);
                if (dr == DialogResult.No)
                {
                    return false;
                }
            }
            return true;
        }

        // check whether the data table has changes
        bool HasChanges()
        {
            var changes = _dtErrors.GetChanges();
            return changes != null && changes.Rows.Count > 0;
        }

        // save changes in the data table
        void SaveChanges()
        {
            try
            {
                // save the changes
                _dataAdapter.Update(_dtErrors);

                // mark table as clean
                _dtErrors.AcceptChanges();

                // disable save changes button until there are more changes
                _btnSaveChanges.Enabled = false;
            }
            catch (Exception x)
            {
                // something went wrong? inform the user
                var dr = MessageBox.Show(this,
                    string.Format(Properties.Resources.FailedToSave, x.Message),
                    Application.ProductName,
                    MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
            }
        }

        #endregion

        //---------------------------------------------------------------------------
        #region ** event handlers

        // confirm that the user wants to discard changes
        protected override void OnClosing(CancelEventArgs e)
        {
            if (!OKToDiscardChanges())
            {
                e.Cancel = true;
            }
            base.OnClosing(e);
        }

        // draw wavy read lines under spelling errors
        void _flex_OwnerDrawCell(object sender, C1.Win.C1FlexGrid.OwnerDrawCellEventArgs e)
        {
            // spell check editable cells (unless we're just measuring)
            if (!e.Measuring && _flex.Cols[e.Col].AllowEditing && e.Row >= _flex.Rows.Fixed)
            {
                var text = _flex.GetDataDisplay(e.Row, e.Col);
                var errors = _spell.CheckText(text);

                if (errors.Count > 0)
                {
                    // draw cell as usual
                    e.Style = _errorStyle;
                    e.DrawCell();

                    // underline errors
                    var ranges = new CharacterRange[errors.Count];
                    for (int i = 0; i < errors.Count; i++)
                    {
                        ranges[i] = new CharacterRange(errors[i].Start, errors[i].Length);
                    }

                    var sf = new StringFormat(e.Style.StringFormat);
                    try
                    {
                        sf.SetMeasurableCharacterRanges(ranges);
                    }
                    catch { }
                    var rc = e.Style.GetTextRectangle(e.Bounds, null);
                    var rgns = e.Graphics.MeasureCharacterRanges(text, e.Style.Font, rc, sf);

                    foreach (var rgn in rgns)
                    {
                        rc = Rectangle.Truncate(rgn.GetBounds(e.Graphics));
                        for (Point pt = new Point(rc.X, rc.Bottom); pt.X + 2 < rc.Right; pt.X += 4)
                        {
                            e.Graphics.DrawLines(Pens.Red, new Point[] 
                            { 
                                new Point(pt.X, pt.Y), 
                                new Point(pt.X + 2, pt.Y - 2), 
                                new Point(pt.X + 4, pt.Y) 
                            });
                        }
                    }
                }
            }
        }

        // enable spell checker in cell editor
        void _flex_SetupEditor(object sender, RowColEventArgs e)
        {
            var tb = _flex.Editor as TextBox;
            if (tb != null)
            {
                _spell.SetActiveSpellChecking(tb, true);
            }
        }

        // toggle column visibility
        void _chkShowAllColumns_CheckedChanged(object sender, EventArgs e)
        {
            for (int i = _flex.Cols.Fixed; i < _flex.Cols.Count; i++)
            {
                _flex.Cols[i].Visible = _chkShowAllColumns.Checked
                    ? true
                    : _flex.Cols[i].AllowEditing;
            }
        }

        // enable save button when there are changes
        void _flex_CellChanged(object sender, RowColEventArgs e)
        {
            _btnSaveChanges.Enabled = HasChanges();
        }

        // cancel dialog
        void _btnCancel_Click(object sender, EventArgs e)
        {
            DialogResult = DialogResult.Cancel;
            Close();
        }

        // save the changes, disable dialog until there are more changes to save
        void _btnSaveChanges_Click(object sender, EventArgs e)
        {
            SaveChanges();
        }

        // perform modal check on the grid
        private void _btnSpell_Click(object sender, EventArgs e)
        {
            var cols = new List<string>();
            foreach (Column col in _flex.Cols)
            {
                if (col.AllowEditing)
                {
                    cols.Add(col.Name);
                }
            }
            var speller = new FlexGridSpellChecker(_flex, cols.ToArray());
            int errors = _spell.CheckControl(speller);
            if (errors == 0)
            {
                MessageBox.Show(this,
                    Properties.Resources.NoSpellingErrors,
                    Application.ProductName);
            }
            else if (errors > 0)
            {
                MessageBox.Show(this,
                    Properties.Resources.SpellCheckingComplete,
                    Application.ProductName);
            }
        }

        // configure spell-checker options
        void _btnOptions_Click(object sender, EventArgs e)
        {
            using (var dlg = new SpellCheckerOptionsDialog())
            {
                dlg.Font = Font;
                dlg.SpellChecker = _spell;
                if (dlg.ShowDialog(this) == DialogResult.OK)
                {
                    // repaint grid to update error indicators
                    _flex.Invalidate();
                }
            }
        }

        #endregion
    }
}
